@@ -1,6 +1,7 @@
 import socketio
 import logging
-import time, datetime
+import time
+import datetime
 import math
 from typing import overload, Union, TypedDict, Literal, Callable, List, Dict, Optional
 
@@ -54,9 +55,11 @@ class DeviceLeftMsg(TypedDict):
     room: str
     device: Device
 
+
 class TimeStampedMsg(TypedDict):
     timeStamp: int
 
+
 class BaseMsg(TimeStampedMsg):
     deviceId: str
     deviceNr: int
@@ -134,6 +137,7 @@ class ErrorMsg(TypedDict):
 def flatten(list_of_lists: List[List]) -> List:
     return [y for x in list_of_lists for y in x]
 
+
 def to_datetime(data: TimeStampedMsg) -> datetime.datetime:
     '''
     extracts the datetime from a data package. if the field `timeStamp` is not present,
@@ -147,6 +151,7 @@ def to_datetime(data: TimeStampedMsg) -> datetime.datetime:
         ts = ts / 1000.0
     return datetime.datetime.fromtimestamp(ts)
 
+
 class Connector:
     __start_time_ns: int = time.time_ns()
     data: Dict[str, List[BaseMsg]] = {}
@@ -159,20 +164,21 @@ class Connector:
     joined_rooms: List[str]
 
     # callback functions
-    on_key: Callable[[KeyMsg], None] = lambda data: None
-    on_pointer: Callable[[PointerMsg], None] = lambda data: None
-    on_acceleration: Callable[[KeyMsg], None] = lambda data: None
-    on_gyro: Callable[[KeyMsg], None] = lambda data: None
-    on_sensor: Callable[[KeyMsg], None] = lambda data: None
+    on_key: Callable[[KeyMsg], None] = None
+    on_pointer: Callable[[PointerMsg], None] = None
+    on_acceleration: Callable[[AccMsg], None] = None
+    on_gyro: Callable[[GyroMsg], None] = None
+    on_sensor: Callable[[Union[GyroMsg, AccMsg]], None] = None
 
-    on_data: Callable[[DataMsg], None] = lambda data: None
-    on_broadcast_data: Callable[[DataMsg], None] = lambda data: None
-    on_all_data: Callable[[List[KeyMsg]], None] = lambda data: None
-    on_device: Callable[[Device], None] = lambda data: None
-    on_devices: Callable[[List[Device]], None] = lambda data: None
-    on_error: Callable[[ErrorMsg], None] = lambda data: None
-    on_room_joined: Callable[[Device], None] = lambda data: None
-    on_room_left: Callable[[Device], None] = lambda data: None
+    on_data: Callable[[DataMsg], None] = None
+    on_broadcast_data: Callable[[DataMsg], None] = None
+    on_all_data: Callable[[List[DataMsg]], None] = None
+    on_device: Callable[[Device], None] = None
+    on_client_device: Callable[[Union[Device, None]], None] = None
+    on_devices: Callable[[List[Device]], None] = None
+    on_error: Callable[[ErrorMsg], None] = None
+    on_room_joined: Callable[[DeviceJoinedMsg], None] = None
+    on_room_left: Callable[[DeviceLeftMsg], None] = None
 
     @property
     def server_url(self):
@@ -197,7 +203,28 @@ class Connector:
         self.joined_rooms = [device_id]
         self.connect()
 
-    def emit(self, event: str, data: BaseSendMsg = {}, broadcast: bool = False):
+    @property
+    def client_device(self):
+        return next((device for device in self.devices if device['isController'] and device['deviceId'] == self.device_id), None)
+
+    def emit(self, event: str, data: BaseSendMsg = {}, broadcast: bool = False, unicast_to: int = None):
+        '''
+        Parameters
+        ----------
+        event : str
+            the event name 
+        
+        data : BaseSendMsg
+            the data to send, fields 'timeStamp' and 'deviceId' are added when they are not present
+
+        Optional
+        --------
+        broadcast : bool
+            wheter to send this message to all connected devices
+
+        unicast_to : int
+            the device number to which this message is sent exclusively. When set, boradcast has no effect. 
+        '''
         if 'timeStamp' not in data:
             data['timeStamp'] = time.time_ns() // 1000000
 
@@ -207,11 +234,19 @@ class Connector:
         if broadcast:
             data['broadcast'] = True
 
+        if type(unicast_to) == int:
+            if 'broadcast' in data:
+                del data['broadcast']
+            data['unicastTo'] = unicast_to
+        
         self.sio.emit(event, data)
 
     def broadcast(self, data: DataMsg):
         self.emit(ADD_NEW_DATA, data=data, broadcast=True)
 
+    def unicast_to(self, data: DataMsg, device_nr: int):
+        self.emit(ADD_NEW_DATA, data=data, unicast_to=device_nr)
+
     def connect(self):
         if self.sio.connected:
             return
@@ -340,7 +375,7 @@ class Connector:
 
         return None
 
-    def set_grid(self, grid: Union[List[str], List[List[str]]], device_id: str = None, device_nr: int = None, broadcast: bool = False):
+    def set_grid(self, grid: Union[List[str], List[List[str]]], device_id: str = None, unicast_to: int = None, broadcast: bool = False):
         '''
         Parameters
         ----------
@@ -350,7 +385,7 @@ class Connector:
         --------
         device_id : str control the device with this id
 
-        device_nr : str control the device with the given number
+        unicast_to : int control the device with the given number
 
         broadcast : bool wheter to send this message to all connected devices
 
@@ -369,14 +404,13 @@ class Connector:
             ADD_NEW_DATA,
             {
                 'type': 'grid',
-                'grid': grid,
-                'deviceId': did,
-                'deviceNr': device_nr
+                'grid': grid
             },
-            broadcast
+            broadcast=broadcast,
+            unicast_to=unicast_to
         )
 
-    def set_color(self, color: str, device_id: str = None, device_nr: int = None, broadcast: bool = False):
+    def set_color(self, color: str, device_id: str = None, unicast_to: int = None, broadcast: bool = False):
         '''
         Parameters
         ----------
@@ -388,7 +422,7 @@ class Connector:
         device_id : str
             control the device with this id
 
-        device_nr : str
+        unicast_to : int
             control the device with the given number
 
         broadcast : bool wheter to send this message to all connected devices
@@ -403,16 +437,14 @@ class Connector:
         set_panel('hsl(0, 100%, 50%)')        # => hsl
         ```
         '''
-        did = self.device_id if device_id is None else device_id
         self.emit(
             ADD_NEW_DATA,
             {
                 'type': 'color',
-                'color': color,
-                'deviceId': did,
-                'deviceNr': device_nr
+                'color': color
             },
-            broadcast
+            broadcast=broadcast,
+            unicast_to=unicast_to
         )
 
     def disconnect(self):
@@ -507,13 +539,19 @@ class Connector:
                 self.on_device(device)
 
     def __on_devices(self, devices: List[Device]):
+        had_client_device = self.client_device is not None
         self.devices = devices
+        if self.on_client_device:
+            has_client_device = self.client_device is not None
+            if (had_client_device and not has_client_device) or (has_client_device and not had_client_device):
+                self.on_client_device(self.client_device)
+
         if self.on_devices is not None:
             self.on_devices(devices)
 
 
 if __name__ == '__main__':
-    logging.basicConfig(level=logging.INFO)
+    logging.basicConfig(level=logging.DEBUG)
     connector = Connector('https://io.lebalz.ch', 'FooBar')
 
     # draw a 3x3 checker board
@@ -523,19 +561,20 @@ if __name__ == '__main__':
         ['black', 'white', 'black'],
     ], broadcast=True)
 
-    connector.on_key = lambda data: print('on key', data)
-    connector.on_broadcast_data = lambda data: print(data)
-    connector.on_data = lambda data: print('on data', data)
-    connector.on_all_data = lambda data: print(data)
-    connector.on_device = lambda data: print('on device', data)
-    connector.on_devices = lambda data: print('on devices', data)
-    connector.on_acceleration = lambda data: print(data)
-    connector.on_gyro = lambda data: print(data)
-    connector.on_sensor = lambda data: None
-    connector.on_room_joined = lambda data: print('room joined: ', data)
-    connector.on_room_left = lambda data: print('room left: ', data)
-    connector.on_pointer = lambda data: print('pointer: ', data)
-    connector.on_error = lambda data: print('error: ', data)
+    connector.on_key = lambda data: logging.info(f'on_key: {data}')
+    connector.on_broadcast_data = lambda data: logging.info(f'on_broadcast_data: {data}')
+    connector.on_data = lambda data: logging.info(f'on_data: {data}')
+    connector.on_all_data = lambda data: logging.info(f'on_all_data: {data}')
+    connector.on_device = lambda data: logging.info(f'on_device: {data}')
+    connector.on_devices = lambda data: logging.info(f'on_devices: {data}')
+    connector.on_acceleration = lambda data: logging.info(f'on_acceleration: {data}')
+    connector.on_gyro = lambda data: logging.info(f'on_gyro: {data}')
+    connector.on_sensor = lambda data: logging.info(f'on_sensor: {data}')
+    connector.on_room_joined = lambda data: logging.info(f'on_room_joined: {data}')
+    connector.on_room_left = lambda data: logging.info(f'on_room_left: {data}')
+    connector.on_pointer = lambda data: logging.info(f'on_pointer: {data}')
+    connector.on_client_device = lambda data: logging.info(f'on_client_device: {data}')
+    connector.on_error = lambda data: logging.info(f'on_error: {data}')    
 
     time.sleep(1)
     print(connector.joined_room_count)
